/*
 *  AAPrimitives.c
 *  glAArg version 0.1.3
 *
 *  Created by arekkusu on Thu Nov 20 2003.
 *  Copyright (c) 2004 arekkusu. All rights reserved.
 *  http://homepage.mac.com/arekkusu/bugs/invariance/
 *
 *  Version history:
 *  0.1.3 Thu Apr  7 2004
 *		* Fix point light defaults. DTS submission for Radeon 7000 mipmap selection problem.
 *  0.1.2 Thu Apr  6 2004
 *		* VAR is now off by default to match regular immediate mode GL. Client must enable it (and handle flushes.)
 *		* Fixed header includes and extern "C" problems.
 *  0.1.1 Wed Mar 31 2004
 *      * Changed glAAPointLight API to use [-1, 1] coordinates clamped inside a unit circle.
 *      * Fixed miplevel brightness for negative falloff values.
 *		* Optimized VAR usage pattern to reduce fence finishing.
 *  0.1   Thu Mar 25 2004
 *  	* Initial public release for iDevGames Vectorized contest.
 *		* Basic points and lines only.
 *
 *  To fix:
 *		buggy line stipple wrap
 *  To implement:
 *		proper multi-ctx support
 *		specifiable size for each texture
 *		line end cap modes
 *		triangles, tristrips, trifans
 *		quads, quad strips
 *		incoming texcoord fixup (requires multitexturing)
 *		3D and transformed 2D projections in screen space
 *		optimize for G3/G4/G5 (altivec, real sqrtf)
 *		optimize for individual GPUs (CVA, ATI rev4, rage128)
 *  To test:
 *		verify that "fastest" hardware mipmap generation actually works on all GPUs
 *		compare G4/G5 frsqrtes error
 *
 *  NOTE: this GL code conditionally uses extensions such as GL_GENERATE_MIPMAP_SGIS when possible.
 *		It should run on any Mac OS X sytem, however 10.2.5 or newer is reccomended for driver stability.
 *
 *  NOTE: compile this with optimizations. These flags are OK:
 *		-Os -mcpu=G3 -mtune=G4 -force_cpusubtype_ALL -ffast-math -frename-registers --param max-gcse-passes=3 -mdynamic-no-pic
 */

#include "AAPrimitives.h"
#include <OpenGL/CGLMacro.h>


// some convienence constants for the texture dimensions:
// point size, size minus one, half size squared, double size, quad size, and point center
#define psz (phf * 2)
#define psm (psz - 1)
#define phs (phf * phf)
#define pdb (psz * 2)
#define pqd (psz * 4)
#define pct (psz + phf)

// VAR convienence constants based on the hardcoded values in the header.
#define vsize (sizeof(Vertex) * VAR_size)
#define line_i ((int)(VAR_size*1.5))


// exposed globals for client app
GLuint					glAA_texture[8];			// the AA sphere texIDs
unsigned char			*glAA_AAtex[8];				// AA texture buffer

// file-scope globals
static int				glAA_mode, glAA_old_mode;   // state globals
static bool				glAA_enable_plight, glAA_enable_stipple;
static Vertex			glAA_v[4];					// temp stack space for three points (quad is biggest primitive)
static int				glAA_vi, glAA_vp, glAA_vcl; // vertex stack index, vertex previous index, vertex color last set
static int				glAA_vc[4];					// vertex color stack usage bools
static float			glAA_alpha_fix;				// for points/lines < 1.0 px wide or long

static float			glAA_point_size;			// point globals
static float			glAA_point_border;
static float			glAA_point_maxr;
static unsigned short	glAA_plight[4];

static float			glAA_line_width;			// line globals
static float			glAA_line_border;
static float			glAA_line_maxr;
static float			glAA_stipple[16], glAA_stipple_phase;
static int				glAA_stipple_count, glAA_stipple_idx;

static Vertex			*glAA_VAR;					// VAR globals
static unsigned short   *glAA_line_index;
static int				glAA_VAR_i, glAA_VAR_skip, glAA_fence;
static GLuint			glAA_fences[VAR_bufs];

static bool				allowMIPGEN, allowVAR;		// control globals
static bool				useMIPGEN, useVAR;
static CGLContextObj	cgl_ctx;					// for immediate mode macros
static float			logtbl[pdb];				// for sphere lookups


#pragma mark -
// private helper functions

// -------------------------------------------------
// set the context for macros
// -------------------------------------------------
static inline void glAASetContext() {
	cgl_ctx = CGLGetCurrentContext();	
}


// -------------------------------------------------
// test hardware mipmap generation
// -------------------------------------------------
bool introspectMIPGEN() {
	bool workingMIPGEN = 0;
	if (allowMIPGEN) {
		// test hardware mipmap generation to verify it does the right thing (some drivers don't)
		GLuint texID[1];
		const unsigned char tex[] = { 0,   0,   0,   0,   0,   0,   0,   0,
									  0,   0,   0,   0,   0,   0,   0,   0,
									  0,   0, 255, 255, 255, 127,   0,   0,
									  0,   0, 255, 255, 255, 255,   0,   0,
									  0,   0, 127, 127, 255, 255,   0,   0,
									  0,   0, 127, 127, 255,   0,   0,   0,
									  0,   0,   0,   0,   0,   0,   0,   0,
									  0,   0,   0,   0,   0,   0,   0,   0 };
		glGenTextures(1, &texID[0]);
		glBindTexture(GL_TEXTURE_2D, texID[0]);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
		// Use hardware filtering and client storage for faster texture generation.
		glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
		glHint(GL_GENERATE_MIPMAP_HINT_SGIS, GL_FASTEST);
		glPixelStorei(GL_UNPACK_CLIENT_STORAGE_APPLE, 1);
		// Generate the full texture mipmap pyramid.
		glTexImage2D(GL_TEXTURE_2D, 0, GL_ALPHA, 8, 8, 0, GL_ALPHA, GL_UNSIGNED_BYTE, tex);
		glPixelStorei(GL_UNPACK_CLIENT_STORAGE_APPLE, 0);
		
		// now check the results
		unsigned char tex1[16];
		glGetTexImage(GL_TEXTURE_2D, 1, GL_ALPHA, GL_UNSIGNED_BYTE, tex1);
		if (((tex1[0] | tex1[1] | tex1[2] | tex1[3] |
		      tex1[4] |                     tex1[7] |
			  tex1[8] |                     tex1[11]|
			  tex1[12]| tex1[13]| tex1[14]| tex1[15]) == 0) &&
			 (tex1[5] == 255) && (tex1[6] == 223) &&
			 (tex1[9] == 127) && (tex1[10]== 191))
			 workingMIPGEN = 1;
		// printf("glAArg: Using hardware mipmap generation: %d\n", workingMIPGEN);
		glDeleteTextures(1, &texID[0]);
	}
//	return workingMIPGEN;
	return 0;			// temporarily always fail, until I figure out what is wrong with GF4MX
}


#pragma mark -
// glAA APIs

// glAAInit/Exit should be called once at program begin/end to set up globals.
// NOTE: this version is NOT multiple-context clean. The client app is responsible for
// setting up GL ctx and state (PixelStore, texture matrix etc)
// THERE MUST BE A VALID CONTEXT CREATED BEFORE CALLING THIS (to set up VAR state etc)
//
void glAAInit() {
	int i;
	for (i=0; i<pdb; i++){							// lookup table has 4x the bitmap resolution
		logtbl[i] = log((float)i/pdb);
	}

	glAASetContext();
	// check renderer capabilities
	const GLubyte *strExt = glGetString(GL_EXTENSIONS);
	
	// this should be true on all renderers except Rage 128 and Radeon 7000.
	allowVAR = (gluCheckExtension ((const unsigned char *)"GL_APPLE_vertex_array_range", strExt));
	useVAR = 0;										// off by default. Client must turn on (and handle flushing.)
	
	// this should be true on any 10.2.3 system or newer (10.2.4 for ATI support).
	allowMIPGEN = (gluCheckExtension ((const unsigned char *)"GL_SGIS_generate_mipmap", strExt));
	useMIPGEN = introspectMIPGEN();					// use it only if it actually works

	// VAR buffers & state
	if (allowVAR) {
		int j;
		if (!glAA_VAR)
			glAA_VAR = (Vertex *) valloc(vsize * VAR_bufs);
		if (!glAA_line_index) {
			glAA_line_index = (unsigned short *)valloc(sizeof(unsigned short)*line_i*VAR_bufs);
			for (j=0; j<VAR_bufs; j++){
				unsigned short v = VAR_size*j;
				for (i=line_i*j; i<line_i*(j+1); i+= 12){
					glAA_line_index[i  ] = v;
					glAA_line_index[i+1] = v+1;
					glAA_line_index[i+2] = v+3;
					glAA_line_index[i+3] = v+2;
					glAA_line_index[i+4] = v+4;
					glAA_line_index[i+5] = v+5;
					glAA_line_index[i+6] = v+7;
					glAA_line_index[i+7 ]= v+6;
					glAA_line_index[i+8] = v+2;
					glAA_line_index[i+9] = v+3;
					glAA_line_index[i+10]= v+5;
					glAA_line_index[i+11]= v+4;
					v=v+8;
				}
			}
		}
		glAA_VAR_i = glAA_fence = 0;

		glEnableClientState(GL_VERTEX_ARRAY_RANGE_APPLE);
		glVertexArrayRangeAPPLE(vsize * VAR_bufs, (GLvoid *)glAA_VAR);
		glGenFencesAPPLE(VAR_bufs, &glAA_fences[0]);

		glEnableClientState(GL_COLOR_ARRAY);
		glColorPointer(4, GL_UNSIGNED_BYTE, sizeof(Vertex), &glAA_VAR[0].rgba);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer(2, GL_FLOAT, sizeof(Vertex), &glAA_VAR[0].tx);
		glEnableClientState(GL_VERTEX_ARRAY);
		glVertexPointer(2, GL_FLOAT, sizeof(Vertex), &glAA_VAR[0].x);
	}

	// setup defaults
	glAA_old_mode = -1;
	glAAPointSize(1.0f);
	glAAPointLight((GLCoord2){0.0f, 0.0f});
	glAALineWidth(1.0f);
	glAALineStipplePhase(1.0f, 0xFFFF, 0.0f);
	glAADisable(GL_LINE_STIPPLE);
	glAAColor1ui(0xFFFFFFFF);
}


void glAAExit() {
	int i;
	glAASetContext();
	if (allowVAR) {
		glVertexArrayRangeAPPLE(0, 0);
		glDisableClientState(GL_VERTEX_ARRAY_RANGE_APPLE);
		free (glAA_VAR); glAA_VAR = nil;
		free (glAA_line_index); glAA_line_index = nil;
		glDeleteFencesAPPLE(VAR_bufs, &glAA_fences[0]);
	}
	for (i=0; i<8; i++){
		if (glAA_texture[i]) { glDeleteTextures(1, &glAA_texture[i]); glAA_texture[i] = 0; }
		if (glAA_AAtex[i])   {free(glAA_AAtex[i]); glAA_AAtex[i] = nil; }
	}
}


// glAA texture functions.
//
// The concept here is simple but the implementation is not. What we want is an AA circle for a point texture
// (lines/tris get strips of the circle stretched over them.) To work with fractional positioning, there must be
// a 1 px empty border around the texture to filter with. Unfortunately, hardware border support isn't reliable.
// Mipmapping further complicates the issue. There are some GL extensions that could make this easier but
// I use the dumbest/least efficient texture layout here so that it works on ALL renderers:
//
// BBBW		B = empty blackness
// BOBW		O = the AA circle
// BBBW		W = wasted texture space (we have to use 2D textures...)
// WWWW			something else (like the font) can go into the wasted space, though.
//
// This layout will retain a 1 px border at every mip level on all hardware.
// For a 32 x 32 circle, this means a 128x128 alpha mip pyramid, or ~ 21k of VRAM.
// For a 128x128 circle, this means a 512x512 alpha mip pyramid, or ~341k of VRAM.
//
static inline float ifun(float x, float y, float F) {   // compute falloff at x,y with exponent F [-1..1]
	float S = (x*x + y);
	float L;
	if (S) {											// estimate sqrt, accurate to about 1/4096
		L = frsqrtes_nr(S);
		L = L * S;
	}
	else 
		L = 0.0f;
	if (F>=0)											// intensity: [-1..0..1] = bloom..normal..fuzzy
		// this is just pow(L/phf, F) but with a lookup table
		S = exp(logtbl[(int)(L*4.0f)]/F);
	else
		// this is a bit more complex-- logarithm ramp up to a solid dot with slight S blend at cusp
		S = L<(1.0f+F)*phf ? 0.0f : exp((1.0f+F*0.9f)/(L/phf-1.0f-F) * logtbl[(int)(L*4.0f)]);
	return 1.0f-S;
}


// 8-way symmetric macro
#define putpixel(x, y,I) {					   \
	unsigned char c = I*255;				   \
	texture[(pct   + x) + (pct   + y)*pqd] =   \
	texture[(pct-1 - x) + (pct   + y)*pqd] =   \
	texture[(pct   + x) + (pct-1 - y)*pqd] =   \
	texture[(pct-1 - x) + (pct-1 - y)*pqd] =   \
	texture[(pct   + y) + (pct   + x)*pqd] =   \
	texture[(pct-1 - y) + (pct   + x)*pqd] =   \
	texture[(pct   + y) + (pct-1 - x)*pqd] =   \
	texture[(pct-1 - y) + (pct-1 - x)*pqd] = c;\
}


void glAAGenerateAATex(float F, GLuint id, bool alias) {			// compute the AA sphere
	if (id < 8) {
		if (!glAA_AAtex[id]) glAA_AAtex[id] = valloc(pqd*pqd);
		unsigned char *texture = glAA_AAtex[id];
		if (texture) {

			int x = phf-1;
			int y = 0;
			int ax;
			#define rs ((phf-1)*(phf-1))
			float T = 0.0f;
			float ys = 0.0f;
				for(ax=y; ax<=x; ax++)
					putpixel(ax, y, ifun(ax, ys, F));
			while (x>y++) {
				ys = (float)y*(float)y;
				float L = sqrt(rs - ys);
				float D = ceilf(L)-L;
				if (D < T) x--;
				putpixel(x, y, (1.0f-D)*(ifun(x, ys, F)));
					for(ax=y; ax<x; ax++)
						putpixel(ax, y, ifun(ax, ys, F));
				T = D;
			}

// TODO: for square line end caps, we'll need to copy the point crossection to a block.

			glAASetContext();
			glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
			if (!glAA_texture[id]) {
				glGenTextures(1, &glAA_texture[id]);
				glBindTexture(GL_TEXTURE_2D, glAA_texture[id]);
				// Ideally we could use GL_CLAMP_TO_BORDER, but this is not available on GF2MX and GF4MX.
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
				// Clamp the max mip level to 4x4 (1 px with 1 px border all around...)
				glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, (int)log2(psz));
				if (useMIPGEN) {
					// Use hardware filtering and client storage for faster texture generation.
					glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
					glHint(GL_GENERATE_MIPMAP_HINT_SGIS, GL_FASTEST);
					glPixelStorei(GL_UNPACK_CLIENT_STORAGE_APPLE, 1);
					// Generate the full texture mipmap pyramid the first time.
					glTexImage2D(GL_TEXTURE_2D, 0, GL_ALPHA, pqd, pqd, 0, GL_ALPHA, GL_UNSIGNED_BYTE, texture);
				}
				else {
					// Generate the entire mip pyramid slowly in software
					gluBuild2DMipmaps(GL_TEXTURE_2D, GL_ALPHA, pqd, pqd, GL_ALPHA, GL_UNSIGNED_BYTE, texture);
				}
			}
			else {
				glBindTexture(GL_TEXTURE_2D, glAA_texture[id]);
				if (useMIPGEN) {
					glPixelStorei(GL_UNPACK_CLIENT_STORAGE_APPLE, 1);
					// Update only the sphere area after the first time.
					glPixelStorei(GL_UNPACK_ROW_LENGTH, pqd);
					glTexSubImage2D(GL_TEXTURE_2D, 0, psz, psz, psz, psz, GL_ALPHA, GL_UNSIGNED_BYTE, texture+pqd*psz+psz);
					glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
				}
				else {
					// Generate the entire mip pyramid slowly in software
					gluBuild2DMipmaps(GL_TEXTURE_2D, GL_ALPHA, pqd, pqd, GL_ALPHA, GL_UNSIGNED_BYTE, texture);
				}
			}
			// Set filtering depending on our aliasing mode.
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, alias?GL_NEAREST:GL_LINEAR);	
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, alias?GL_NEAREST:GL_LINEAR_MIPMAP_LINEAR);
			// now fix the upper mip levels so that 1 px points/lines stay bright:
			// at 1 px, the circle texture's default brightness is pi/4 = 0.78. This is geometrically
			// correct, but for usability reasons we want 1 px wide lines to display at full 1.0 brightness.
			// the following is an approximation sampling of the texture to do that, getting values that
			// "look ok" for exponential falloff in either direction.
			static unsigned char mipfix[4];
			unsigned char fix = (texture[(int)(psz+phf*0.2f)+pct*pqd]+texture[(int)(psz+phf*0.7f)+pct*pqd])>>1;
			mipfix[0]=mipfix[1]=mipfix[2]=mipfix[3]=fix;
			glTexSubImage2D(GL_TEXTURE_2D, (int)log2(phf), 2, 2, 2, 2, GL_ALPHA, GL_UNSIGNED_BYTE, mipfix);
			glTexSubImage2D(GL_TEXTURE_2D, (int)log2(psz), 1, 1, 1, 1, GL_ALPHA, GL_UNSIGNED_BYTE, mipfix);
			if (useMIPGEN) {
				glPixelStorei(GL_UNPACK_CLIENT_STORAGE_APPLE, 0);
			}
		}
	}
}


// glAA APIs mirroring OpenGL functionality
//
// The concept here is very simple, just buffer incoming vertex data until the current primitive is full,
// then calculate appropriate textured triangles for the primitive and submit in a group.
//
void glAABegin(GLenum mode) {			// cgl_ctx must have been set prior to this. so call glAASetContext() when needed.
	glAA_vi = glAA_vp = 0;
	if (glAA_old_mode != mode) {
		switch (mode) {
			case GL_POINTS: {
				glAA_alpha_fix = MIN(1.0f, glAA_point_size);
				// we always have 4 vertex alignment
			} break;
			case GL_LINE_LOOP:
			case GL_LINE_STRIP:
			case GL_LINES: {
				glAA_alpha_fix = MIN(1.0f, glAA_line_width);
				// force 8 vertex alignment
				glAA_VAR_skip = glAA_VAR_i&4;				
			} break;
		}
		glAAFlush();
	}
	glAA_mode = mode;
}


void glAAEnd() {
	glAA_old_mode = glAA_mode;
	if ((glAA_mode == GL_LINE_STRIP) || (glAA_mode == GL_LINE_LOOP)) {
		if (glAA_mode == GL_LINE_LOOP) glAAVertex2f(glAA_v[3].x, glAA_v[3].y);
		glAA_vc[glAA_vp] = 0;
	}
	glAA_vi = glAA_vp = 0;
	// reset stipple index
	glAA_stipple_idx = 0;
}


void glAAFlush() {
	if (useVAR){
	static int glAA_VAR_ip = 0;
		int length = glAA_VAR_i-glAA_VAR_ip;
				
		if (length > 0){
			switch (glAA_mode) {
				case GL_POINTS: {
		//		if (glAA_VAR_ip &3) printf("glAArg: WARNING! flushing pnt  from bad align %x %d\n", glAA_VAR_ip, glAA_VAR_skip);
		//		printf("glAArg: flushing %4d pnts  (%5d vertices) from %04x\n", length/4, length, glAA_VAR_ip);
					glFlushVertexArrayRangeAPPLE(length*sizeof(Vertex), &glAA_VAR[glAA_VAR_ip]);
					glDrawArrays(GL_QUADS, glAA_VAR_ip, length);
				} break;
				case GL_LINE_LOOP:
				case GL_LINE_STRIP:
				case GL_LINES: {
		//		if (glAA_VAR_ip &7) printf("glAArg: WARNING! flushing line from bad align %x %d\n", glAA_VAR_ip, glAA_VAR_skip);
		//		printf("glAArg: flushing %4d lines (%5d vertices) from %04x\n", length/8, length, glAA_VAR_ip);
					glFlushVertexArrayRangeAPPLE(length*sizeof(Vertex), &glAA_VAR[glAA_VAR_ip]);
		//		printf("drawing %5d line elements from %04x\n", length+(length>>1), glAA_VAR_ip+(glAA_VAR_ip>>1));
					glDrawElements(GL_QUADS, length+(length>>1), GL_UNSIGNED_SHORT, &glAA_line_index[glAA_VAR_ip+(glAA_VAR_ip>>1)]);
				} break;
			}
		}
		if (glAA_VAR_skip) {
			glAA_VAR_i += glAA_VAR_skip;
			glAA_VAR_ip += glAA_VAR_skip;
			glAA_VAR_skip = 0;
		}

		if (glAA_VAR_i && !(glAA_VAR_i%VAR_size)){
			glSetFenceAPPLE(glAA_fences[glAA_fence]);
			glAA_fence++; if (glAA_fence >= VAR_bufs) glAA_fence = 0;
			glAA_VAR_i = glAA_VAR_ip = VAR_size*glAA_fence;
		//  printf("glAArg: VAR index fenced to %04x\n", glAA_VAR_i);
			glFinishFenceAPPLE(glAA_fences[glAA_fence]);
		}
		else glAA_VAR_ip = glAA_VAR_i;
	}
	else {			// immediate mode, do nothing.
//		glFlush();
	}
}


void glAAEnable(GLenum cap) {
	if      (cap == GL_LINE_STIPPLE) glAA_enable_stipple = 1;
	else if (cap == GL_VERTEX_ARRAY_RANGE_APPLE) useVAR = allowVAR;
	else glEnable(cap);
}


void glAADisable(GLenum cap) {
	if      (cap == GL_LINE_STIPPLE) glAA_enable_stipple = 0;
	else if (cap == GL_VERTEX_ARRAY_RANGE_APPLE) useVAR = 0;
	else glDisable(cap);
}


void glAAPointSize(float s) {
	float size = ((s>0.0f) ? s: 1.0f);
	
	if (size != glAA_point_size) {
		float r = MAX(1.0f, size);
		// find min/max using float trick (faster than floor/ceil or casts, only works for positive IEEE 32 bit floats)
		float minr = (r+8388608.f)-8388608.f;
		if (r < minr) minr -= 1.0f;
		float maxr = (minr == r) ? minr : minr+1.0f;
		// calculate the tex coords for a 0.5px outset, doing two divides at once
		float den2 = 1.0f/(maxr*minr);
		float ratio = phf*minr*den2;
		glAA_point_border = ratio+(phf*maxr*den2*1.5f-ratio)*(maxr-r);
		glAA_point_maxr = maxr;
		glAA_point_size = size;
	}
	if (glAA_mode == GL_POINTS)							// allow changing before glAAEnd()
		glAA_alpha_fix = MIN(1.0f, glAA_point_size);
}


void glAAPointLight(GLCoord2 light) {   // passed in point normalized to -1..1
	float x = light.x, y = light.y;
	// contain coords within circle to avoid interpolation artifacts
	float factor = x*x+y*y;
	if (factor) {
		float perpd = frsqrtes_nr(factor);
		if (perpd < 1){
			x *= perpd;
			y *= perpd;
		}
	}
	x+=1.0f; y+=1.0f;
	
	glAA_plight[0] = 256*((2.0f-x)*(2.0f-y));
	glAA_plight[1] = 256*(      x *(2.0f-y));
	glAA_plight[2] = 256*(      x *      y);
	glAA_plight[3] = 256*((2.0f-x)*      y);
	glAA_enable_plight = (light.x || light.y);
}


void glAALineWidth(float w) {
	float width = ((w>0.0f) ? w : 1.0f);
	
	if (width != glAA_line_width) {
		float r = MAX(1.0f, width);
		// find min/max using float trick (faster than floor/ceil or casts, only works for positive IEEE 32 bit floats)
		float minr = (r+8388608.f)-8388608.f;
		if (r < minr) minr -= 1.0f;
		float maxr = (minr == r) ? minr : minr+1.0f;
		// calculate the tex coords for a 0.5px outset, doing two divides at once
		float den2 = 1.0f/(maxr*minr);
		float ratio = phf*minr*den2;
		glAA_line_border = ratio+(phf*maxr*den2*1.5f-ratio)*(maxr-r);
		glAA_line_maxr = maxr;
		glAA_line_width = width;
	}
	if ((glAA_mode == GL_LINES) || (glAA_mode == GL_LINE_LOOP) || (glAA_mode == GL_LINE_STRIP))	// allow changing before glAAEnd()
		glAA_alpha_fix = MIN(1.0f, width);
}


inline void glAALineStipple (GLint factor, GLushort pattern) {
	glAALineStipplePhase((GLfloat)factor, pattern, 0.0f);
}


// stipple pattern is buggy. wrapping across segments and phase adjustment are broken.
void glAALineStipplePhase (GLfloat factor, GLushort pattern, GLfloat phase) {
	float fac = MAX(1.0f, factor), len = 0.0f;
	int b, i=0, pat = pattern, m=pat&1;
	for (b=0; b<16;){
		int p=pat&1;
		while ((m==p) && (b<16)){
			len += fac;
			pat >>= 1;
			p=pat&1;
			b++;
		}
		glAA_stipple[i++] = m?len:-len;
		m = p;
		len = 0.0f;
	}
	if (i==1){
		glAA_stipple[0] = FLT_MAX;
	}
	else if (i&1) {				// if odd number of dashes then we wrap, so adjust the phase
//		glAA_stipple[i+1] = 0.0f;
		i--;
		phase += fabsf(glAA_stipple[i]);
		glAA_stipple[0] += glAA_stipple[i];
	}
	glAA_stipple_phase = phase;
	glAA_stipple_count = i;
//for (b=0; b<glAA_stipple_count; b++){ printf("%.0f ", glAA_stipple[b]); }
//printf("  phase %.0f\n", glAA_stipple_phase);
}


inline void glAAColor3f(float R, float G, float B){
	glAAColor4f(R, G, B, 1.0f);
}


void glAAColor4f(float R, float G, float B, float A){
	glAA_vc[glAA_vi] = 1;
	glAA_vcl = glAA_vi;
	glAA_v[glAA_vi].rgba = ((int)(R*255.0f)<<24) | ((int)(G*255.0f)<<16) | ((int)(B*255.0f)<<8) | (int)(A*255.0f);
}


void glAAColor4ubv(const GLubyte *c){
	if (c) {
		glAA_vc[glAA_vi] = 1;
		glAA_vcl = glAA_vi;
		glAA_v[glAA_vi].rgba = *(GLuint *)c;
	}
}


void glAAColor1ui(GLuint c){
	glAA_vc[glAA_vi] = 1;
	glAA_vcl = glAA_vi;
	glAA_v[glAA_vi].rgba = c;
}


// here's where all the real work is done.
void glAAVertex2f(float x, float y) {
	switch (glAA_mode) {
		case GL_POINTS: {
			float bord = glAA_point_border;			// shadow globals
			float maxr = glAA_point_maxr;
			float alpha_fix = glAA_alpha_fix;

			float size = maxr+1.0f;
			float centx = x-(maxr+1.0f)*0.5f;
			float centy = y-(maxr+1.0f)*0.5f;

			// adjust vertex alpha if needed
			int c1 = glAA_v[glAA_vcl].rgba;
			if (alpha_fix < 1.0f) {
				int int_alpha_fix = alpha_fix * 255.0f;
				c1 = (c1 & 0xFFFFFF00) | (((c1 & 0x000000FF)*int_alpha_fix)>>8);
			}
			int c2 = c1, c3 = c1, c4 = c1;
			
			// apply saturation lighting if needed
			if (glAA_enable_plight) {
				c1 = scalesatRGB(c1, glAA_plight[0]);
				c2 = scalesatRGB(c2, glAA_plight[1]);
				c3 = scalesatRGB(c3, glAA_plight[2]);
				c4 = scalesatRGB(c4, glAA_plight[3]);
			}
			
			if (useVAR){
				int i = glAA_VAR_i;
				glAA_VAR[i  ].rgba = c1;
				glAA_VAR[i+1].rgba = c2;
				glAA_VAR[i+2].rgba = c3;
				glAA_VAR[i+3].rgba = c4;
				glAA_VAR[i  ].tx = glAA_VAR[i+3].tx =
				glAA_VAR[i  ].ty = glAA_VAR[i+1].ty = psz-bord;
				glAA_VAR[i+1].tx = glAA_VAR[i+2].tx = 
				glAA_VAR[i+2].ty = glAA_VAR[i+3].ty = pdb+bord;
				glAA_VAR[i  ].x = glAA_VAR[i+3].x = centx;
				glAA_VAR[i  ].y = glAA_VAR[i+1].y = centy;
				glAA_VAR[i+1].x = glAA_VAR[i+2].x = centx+size;
				glAA_VAR[i+2].y = glAA_VAR[i+3].y = centy+size;
				glAA_VAR_i += 4; 
				int room = VAR_size*(glAA_fence+1)-glAA_VAR_i;
				if (room < 4) { glAA_VAR_skip = room; glAAFlush();}
			}
			else {	// immediate mode
				glBegin(GL_QUADS);
					glColor4ubv((const GLubyte *)&c1);
					glTexCoord2f(psz-bord, psz-bord);
					glVertex2f(centx,      centy);
					glColor4ubv((const GLubyte *)&c2);
					glTexCoord2f(pdb+bord, psz-bord);
					glVertex2f(centx+size, centy);
					glColor4ubv((const GLubyte *)&c3);
					glTexCoord2f(pdb+bord, pdb+bord);
					glVertex2f(centx+size, centy+size);
					glColor4ubv((const GLubyte *)&c4);
					glTexCoord2f(psz-bord, pdb+bord);
					glVertex2f(centx,      centy+size);		
				glEnd();
			}
		} break;

		case GL_LINE_LOOP:
		case GL_LINE_STRIP:
		case GL_LINES: {
			if (glAA_vi < 1) {
				glAA_v[glAA_vi].x = x; glAA_v[glAA_vi++].y = y;
				if (glAA_mode == GL_LINE_LOOP) {	// also remember the first point for looping
					glAA_v[3].x = x; glAA_v[3].y = y;
				}
			} else {
				if (glAA_mode != GL_LINES){			// remember the new point
					glAA_v[glAA_vi].x = x; glAA_v[glAA_vi].y = y;
				}
				float x1 = glAA_v[glAA_vp].x;		// shadow globals
				float y1 = glAA_v[glAA_vp].y;
				float bord = glAA_line_border;
				float maxr = glAA_line_maxr;				
				float alpha_fix = glAA_alpha_fix;
				bool enable_stipple = glAA_enable_stipple;

				float perp_y = x1-x;
				float perp_x = y-y1;
				float perpd, leng;
			
				float factor = perp_y*perp_y+perp_x*perp_x;
				if (factor) {
					perpd = frsqrtes_nr(factor);
					perp_y *= perpd;				// normalize to 1px
					perp_x *= perpd;
					leng = __fres(perpd);			// to get the actual sqrt, normally do perpd*factor. but in this case the
													// error tends to make the estimate too low, and we drop stipple segments.
													// suprisingly, another estimate with __fres give a closer approximation,
													// but there are still subpixel errors in long lines...
					alpha_fix *= MIN(1.0f, leng);
					leng = MAX(1.0f, leng);
					perpd= MIN(1.0f, perpd);
//					x += perp_y*0.5f;
//					y -= perp_x*0.5f;
					x1 -= perp_y*0.5f;
					y1 += perp_x*0.5f;
				} else {
					perp_y = 0.0f;
					perp_x = 1.0f;
					perpd = leng = 1.0f;
				}

				float width = (maxr+1.0f)*0.5f;
				float unit_x = -perp_y;
				float unit_y = perp_x;
				perp_y *= width;
				perp_x *= width;
				float parl_x = -perp_y;
				float parl_y =  perp_x;

				float s = 0.0f;
				int b = glAA_stipple_idx;
//		printf("start idx %d   ", b);
				int i = glAA_VAR_i;
				const float lengm1 = leng-1.0f;
				float dash = FLT_MAX, gap = 0;
				while (s <= lengm1) {
				
					if (enable_stipple){						// this chunk is buggy and in progress
						if (s==0.0f) {
							dash = glAA_stipple[b];
							if (dash>0.0f) {
								dash -= glAA_stipple_phase;
								b = (b+1)%glAA_stipple_count;
								gap = -glAA_stipple[b];
							}
							else {
								s-=dash + glAA_stipple_phase;
								x1 -= unit_x*(dash + glAA_stipple_phase);
								y1 -= unit_y*(dash + glAA_stipple_phase);
								b = (b+1)%glAA_stipple_count;
								dash = glAA_stipple[b];
								b = (b+1)%glAA_stipple_count;
								gap = -glAA_stipple[b];
							}
						}
						else {
							x1 += unit_x*(dash+gap);
							y1 += unit_y*(dash+gap);
							dash = fabsf(glAA_stipple[b]);
							b = (b+1)%glAA_stipple_count;
							gap =  fabsf(glAA_stipple[b]);
						}
						b = (b+1)%glAA_stipple_count;
					}
					
					if (s >= leng-dash) {dash = MAX(1.0f, leng-s);}		// clip final dash
//					printf("glAArg: line w %.1f leng %.1f using(%d) %.1f dash [%d], s is %.1f, gap is %.1f [%d]\n", width, leng, glAA_enable_stipple, dash, b, s, gap, b+1);
					x = x1 + (dash-1.0f)*unit_x;
					y = y1 + (dash-1.0f)*unit_y;

					// find stipple weights
					float cw2 = MIN(1.0f, (s+0.5f)*perpd);
					float cp2 = MAX(0.0f, (s-width)*perpd);
					float cw4 = MIN(1.0f, (s-0.5f+dash)*perpd);
					float cp4 = MIN(1.0f, (s+dash+width)*perpd);
			
					// pick vertex colors and adjust alpha if needed
					int c1 = (glAA_vc[glAA_vp]) ? glAA_v[glAA_vp].rgba : glAA_v[glAA_vcl].rgba;
					int c2 = (glAA_vc[glAA_vi]) ? glAA_v[glAA_vi].rgba : glAA_v[glAA_vcl].rgba;
					if (alpha_fix < 1.0f) {
						int int_alpha_fix = alpha_fix * 255.0f;
						c1 = (c1 & 0xFFFFFF00) | (((c1 & 0x000000FF)*int_alpha_fix)>>8);
						c2 = (c2 & 0xFFFFFF00) | (((c2 & 0x000000FF)*int_alpha_fix)>>8);
					}

					if (useVAR){
						// should use altivec, here...
						glAA_VAR[i  ].rgba = glAA_VAR[i+1].rgba = lerpRGBA(c1, c2, cp2);
						glAA_VAR[i+2].rgba = glAA_VAR[i+3].rgba = lerpRGBA(c1, c2, cw2);
						glAA_VAR[i+4].rgba = glAA_VAR[i+5].rgba = lerpRGBA(c1, c2, cw4);
						glAA_VAR[i+6].rgba = glAA_VAR[i+7].rgba = lerpRGBA(c1, c2, cp4);

						glAA_VAR[i+3].tx = glAA_VAR[i+1].tx = glAA_VAR[i+5].tx = glAA_VAR[i+7].tx =
						glAA_VAR[i  ].ty = glAA_VAR[i+1].ty = glAA_VAR[i+6].ty = glAA_VAR[i+7].ty = psz-bord;
						glAA_VAR[i+2].tx = glAA_VAR[i  ].tx = glAA_VAR[i+4].tx = glAA_VAR[i+6].tx = pdb+bord;
						glAA_VAR[i+3].ty = glAA_VAR[i+2].ty = glAA_VAR[i+5].ty = glAA_VAR[i+4].ty = pct;
						
						glAA_VAR[i  ].x = x1-perp_x-parl_x;
						glAA_VAR[i+1].x = x1+perp_x-parl_x;
						glAA_VAR[i+2].x = x1-perp_x;
						glAA_VAR[i+3].x = x1+perp_x;
						glAA_VAR[i+4].x = x-perp_x;
						glAA_VAR[i+5].x = x+perp_x;
						glAA_VAR[i+6].x = x-perp_x+parl_x;
						glAA_VAR[i+7].x = x+perp_x+parl_x;
			
						glAA_VAR[i  ].y = y1-perp_y-parl_y;
						glAA_VAR[i+1].y = y1+perp_y-parl_y;
						glAA_VAR[i+2].y = y1-perp_y;
						glAA_VAR[i+3].y = y1+perp_y;
						glAA_VAR[i+4].y = y-perp_y;
						glAA_VAR[i+5].y = y+perp_y;
						glAA_VAR[i+6].y = y-perp_y+parl_y;
						glAA_VAR[i+7].y = y+perp_y+parl_y;
			
						glAA_VAR_i += 8;
						int room = VAR_size*(glAA_fence+1)-glAA_VAR_i;
						if (room < 8) { glAA_VAR_skip = room; glAAFlush();}

						i = glAA_VAR_i;
					}
					else { // immediate	
						int cl1 = lerpRGBA(c1, c2, cp2);
						int cl2 = lerpRGBA(c1, c2, cw2);
						int cl3 = lerpRGBA(c1, c2, cw4);
						int cl4 = lerpRGBA(c1, c2, cp4);
						glBegin(GL_QUAD_STRIP);
							glColor4ubv((const GLubyte *)&cl1);
							glTexCoord2f(pdb+bord,  psz-bord);
							glVertex2f(x1-perp_x-parl_x, y1-perp_y-parl_y);
							glTexCoord2f(psz-bord,  psz-bord);
							glVertex2f(x1+perp_x-parl_x, y1+perp_y-parl_y);

							glColor4ubv((const GLubyte *)&cl2);
							glTexCoord2f(pdb+bord,  pct);
							glVertex2f(x1-perp_x, y1-perp_y);
							glTexCoord2f(psz-bord,  pct);
							glVertex2f(x1+perp_x, y1+perp_y);
						
							glColor4ubv((const GLubyte *)&cl3);				
							glTexCoord2f(pdb+bord,  pct);
							glVertex2f(x-perp_x, y-perp_y);
							glTexCoord2f(psz-bord,  pct);
							glVertex2f(x+perp_x, y+perp_y);

							glColor4ubv((const GLubyte *)&cl4);
							glTexCoord2f(pdb+bord,  psz-bord);
							glVertex2f(x-perp_x+parl_x, y-perp_y+parl_y);
							glTexCoord2f(psz-bord,  psz-bord);
							glVertex2f(x+perp_x+parl_x, y+perp_y+parl_y);
						glEnd();
					}
					s  += dash+gap;
				}
				glAA_stipple_idx = b;

			//	printf("glAArg: add line at %d alphafix is %f\n", glAA_VAR_i, glAA_alpha_fix);

				if (glAA_mode == GL_LINES) {
					glAA_vi = glAA_vc[0] = glAA_vc[1] = 0;
				}
				else {
					glAA_vc[glAA_vp] = 0;		// note, the other vertex color is cleaned in glEnd,
					glAA_vp = glAA_vi;			// so that interpolation works between strip vertices
					glAA_vi++; if (glAA_vi == 3) glAA_vi = 1;
				}
			}
		} break;
	}
}
